#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/wait.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/mount.h>

#define FALSE (0)
#define TRUE (1)


typedef struct _filelist {
	char *filename;
	struct _filelist *next;
} filelist;


#define SORTPUSH
#ifdef SORTPUSH

int push(filelist **list, char *filename) {
	filelist *prev;
	filelist *this;
	filelist *listp;
	int done = FALSE;

	if (*list == NULL) {
		printf("list initialization failed\n");
		return -1;
	} else {
		listp = (*list);
		prev = NULL;
		while (!done && listp->filename != NULL) {
			if (strcmp(filename, listp->filename) < 0) {  /* new name is less */
				this = malloc(sizeof(filelist));
				if (this == NULL) {
					perror("malloc");
					return -1;
				}
				this->filename = filename;
				this->next = listp;
				if (prev == NULL) {
					(*list) = this;
				} else {
					prev->next = this;
				}
				done = TRUE;
			}
			prev = listp;
			listp = listp->next;
		}
		if (!done) {
			listp->filename = filename;
			listp->next = malloc(sizeof(filelist));
			listp = listp->next;
			if (listp == NULL) {
				perror("malloc");
				return -1;
			}
			listp->filename = NULL;
			listp->next = NULL;
		}
		return 0;
	}
}

#else

int push(filelist **list, char *filename) {
	if (*list == NULL) {
		printf("list initialization failed\n");
		return -1;
	} else {
		(*list)->filename = filename;
		(*list)->next = malloc(sizeof(filelist));
		(*list) = (*list)->next;
		(*list)->filename = NULL;
		(*list)->next = NULL;
		return 0;
	}
}

#endif /*SORTPUSH*/


int waitformounttime(void) {
	struct stat statbuf;
	int i=60;

	while ((stat("/empeg/var/config.ini", &statbuf) != 0) && (i>0)) {
		if ((errno != ENOENT) && (errno != ENOTDIR)) {
			perror("stat");
			continue;
		}
		i--;
		sleep(1);
	}
	if (i<=0) {
		printf("empeg-preinit: config.ini read timed out.  mount-time scripts not executed.\n");
		return 0;
	}

	return -1;

}


void freescripts(filelist *list) {
	filelist *old;

	while (list != NULL) {
		free(list->filename);
		old = list;
		list = list->next;
		free(old);
	}

}


void execscripts(filelist *list, char *preinitd) {
	int namelen;
	char *name;
	filelist *old;
	struct stat statbuf;

	while (list != NULL) {
		if (list->filename != NULL) {
			namelen = strlen(preinitd) + strlen(list->filename) + 2; /* 2 = / and \0 */
			if ((name = malloc(namelen)) != NULL) {
				snprintf(name, namelen, "%s/%s", preinitd, list->filename);
				if (stat(name, &statbuf) == 0) {
					free(name);
					if (statbuf.st_mode & S_IFREG) {
						if (statbuf.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)) {
							/* XXX */
							namelen += 8;
							if ((name = malloc(namelen)) != NULL) {
								snprintf(name, namelen, "/bin/sh %s/%s", preinitd, list->filename);
								system(name);
								free(name);
							} else {
								perror("malloc");
								printf("cannot execute %s\n", list->filename);
							}
							/* XXX - Section from XXX above to here may be replaceable by:
							system(name);
							   and moving the free(name) above to below, which would make
							   non-script executables work, but it failed on me before */
						} else {
							printf("%s is not marked as executable.  Not executing.\n", list->filename);
						}
					} else {
						printf("%s is not a regular file.  Not executing.\n", list->filename);
					}
				} else {
					perror("stat");
					printf("%s doesn't exist?\n", list->filename);
					free(name);
				}
			} else {
				perror("malloc");
				printf("cannot execute %s\n", list->filename);
			}
			free(list->filename);
			old = list;
			list = list->next;
			free(old);
		} else {
			old = list;
			list = list->next;
			free(old);
		}
	}

}


void timeout(int i) {
	printf("empeg-preinit has timed out\n");
	printf("  You should solve this problem ASAP\n");
	printf("  continuing...\n");
	execl("/sbin/init", "init", NULL);
}


int main(int argc, char *argv[]) {
	struct stat statbuf;
	char *preinitd;
	DIR *dir;
	struct dirent *file;
	char *name;
	int namelen;
	pid_t pid;

	filelist *now, *block, *mounttime;
#ifndef SORTPUSH
	filelist *nowp, *blockp, *mounttimep;
#endif

	signal(SIGALRM, timeout);
	alarm(60);

	printf("\nempeg-preinit starting...\n"); /* leading \n because the empeg doesn't provide one at this point */

	now=malloc(sizeof(filelist));
#ifndef SORTPUSH
	nowp=now;
#endif
	if (now == NULL) {
		perror("malloc");
		goto init;
	}
	now->filename = NULL;
	now->next = NULL;

	block=malloc(sizeof(filelist));
#ifndef SORTPUSH
	blockp=block;
#endif
	if (block == NULL) {
		perror("malloc");
		free(now);
		goto init;
	}
	block->filename = NULL;
	block->next = NULL;

	mounttime=malloc(sizeof(filelist));
#ifndef SORTPUSH
	mounttimep=mounttime;
#endif
	if (mounttime == NULL) {
		perror("malloc");
		free(now);
		free(block);
		goto init;
	}
	mounttime->filename = NULL;
	mounttime->next = NULL;

	preinitd = NULL;

#if 0 /* Doesn't seem to work.  Creates a stack dump.  May be because the
				 transistion from flash root to hard drive root is not yet completed */
	/* Attempt to mount /programs0 */
	if (stat("/programs0", &statbuf) == 0) {
		if (!(statbuf.st_mode & S_IFDIR)) {
			printf("/programs0 exists, but is not a directory\n");
			preinitd = strdup("/etc/preinit.d");
		}
	} else { /* /programs0 doesn't exist */
		if (mount("/dev/root", "/", "ext2", MS_MGC_VAL|MS_REMOUNT, "") == 0) {
			if (mkdir("/programs0", 0777) != 0) {
				printf("/programs0 does not exist but cannot be created\n");
				preinitd = strdup("/etc/preinit.d");
			}
			/* create /programs1, too, while / mounted rw */
			if (stat("/programs1", &statbuf) == 0) {
				if (!(statbuf.st_mode & S_IFDIR)) {
					printf("/programs1 exists, but is not a directory\n");
				}
			} else {
				if (mkdir("/programs1", 0777) != 0) {
					printf("/programs1 does not exist but cannot be created\n");
				}
			}
			if (mount("/dev/hda5", "/", "ext2", MS_MGC_VAL|MS_REMOUNT|MS_RDONLY, "") != 0) {
				printf("WARNING: cannot remount / read-only!!!\n");
			}
		} else {
			printf("cannot remount / read-write to create /programs0\n");
		}
	}
#endif

	if (preinitd == NULL) { /* if it's already been set, the mount won't work */
		if (mount("/dev/hda2", "/programs0", "ext2", MS_MGC_VAL|MS_RDONLY, "") == 0) {
			if (stat("/programs0/preinit.d", &statbuf) == 0) {
				if (statbuf.st_mode & S_IFDIR) {
					preinitd = strdup("/programs0/preinit.d");
				} else {
					printf("/programs0/preinit.d exists, but is not a directory\n");
					preinitd = strdup("/etc/preinit.d");
				}
			} else {
				printf("/programs0/preinit.d doesn't exist\nYou might want to consider placing your startup scripts there\n");
				preinitd = strdup("/etc/preinit.d");
			}
		} else {
			perror("mount: /programs0");
			if (errno == ENOTDIR) {
				printf("/programs0 mount point doesn't exist\n");
			} else if (errno == EINVAL) {
				printf("/dev/hda2 cannot be mounted.\nYou might want to consider creating a filesystem on it.\n");
			}
			preinitd = strdup("/etc/preinit.d");
		}
	}

	/* Might as well try to mount /programs1, too */
	if (mount("/dev/hdc2", "/programs1", "ext2", MS_MGC_VAL|MS_RDONLY, "") != 0) {
		if (errno != ENXIO) {
			perror("mount: /programs1");
		}
		if (errno == ENOTDIR) {
			printf("/programs1 mount point doesn't exist\n");
		} else if (errno == ENOTBLK) { /* ENOTBLK means that the device doesn't exist */
			printf("/dev/hdc2 cannot be mounted.\nYou might want to consider creating a filesystem on it.\n");
		}
	}

	printf("Using %s as startup script directory\n", preinitd);

	if (stat(preinitd, &statbuf) == 0) {
		if (statbuf.st_mode & S_IFDIR) {
			if ((dir = opendir(preinitd)) != NULL) {
				while ((file = readdir(dir)) != NULL) {
					if (!((strcmp(file->d_name, ".")==0) || (strcmp(file->d_name, "..")==0))) {
						switch (file->d_name[0]) {
							case 'N':
								namelen = strlen(file->d_name) + 1;
								if ((name = malloc(namelen)) != NULL) {
									snprintf(name, namelen, "%s", file->d_name);
#ifdef SORTPUSH
									push(&now, name);
#else
									push(&nowp, name);
#endif /* SORTPUSH */
								} else {
									perror("malloc");
								}
								break;

							case 'B':
								namelen = strlen(file->d_name) + 1;
								if ((name = malloc(namelen)) != NULL) {
									snprintf(name, namelen, "%s", file->d_name);
#ifdef SORTPUSH
									push(&block, name);
#else
									push(&blockp, name);
#endif /* SORTPUSH */
								} else {
									perror("malloc");
								}
								break;

							case 'M':
								namelen = strlen(file->d_name) + 1;
								if ((name = malloc(namelen)) != NULL) {
									snprintf(name, namelen, "%s", file->d_name);
#ifdef SORTPUSH
									push(&mounttime, name);
#else
									push(&mounttimep, name);
#endif /* SORTPUSH */
								} else {
									perror("malloc");
								}
								break;

							default:
								printf("empeg-preinit: %s: type '%c' not defined\n", file->d_name, file->d_name[0]);
								break;
						}
					}
				}
				closedir(dir);

				execscripts(now, preinitd);

				if (mounttime->filename != NULL) {
					pid = fork();
					if (pid == 0) {
						signal(SIGALRM, SIG_DFL);
						if (waitformounttime()) {
							execscripts(mounttime, preinitd);
						} else {
							freescripts(mounttime);
						}
					} else if (pid == -1) {
						perror("fork");
						printf("mount time scripts will not be executed\n");
					}
				} else {
					pid = 1;
				}

				if (block->filename != NULL) {
					signal(SIGALRM, SIG_DFL);
					execscripts(block, preinitd);
				}
				signal(SIGALRM, timeout); /* In case the script does not block and something else goes wrong */
			}
		}
	} else {
		printf("no %s\n", preinitd);
		pid = 1;
	}

	free(preinitd);

	if (pid > 0) {
init:
		printf("empeg-preinit complete\n");

		execl("/sbin/init", "init", NULL);

		return -1;
	} else {
		return 0;
	}

}

/* vim:set tabstop=2 shiftwidth=2 cindent: */
